package kc.mega;
import kc.mega.game.BattleField;
import kc.mega.game.GameState;
import kc.mega.gun.Gun;
import kc.mega.movement.Movement;
import kc.mega.other.DatasetWriter;
import kc.mega.other.Painter;
import kc.mega.other.Radar;
import kc.mega.other.Shielder;
import kc.mega.other.Stopwatch;
import kc.mega.other.Strategy;
import kc.mega.wave.WaveManager;

import java.awt.Color;
import java.awt.Graphics2D;

import robocode.AdvancedRobot;
import robocode.Bullet;
import robocode.BulletHitBulletEvent;
import robocode.BulletHitEvent;
import robocode.DeathEvent;
import robocode.Event;
import robocode.HitByBulletEvent;
import robocode.HitRobotEvent;
import robocode.ScannedRobotEvent;
import robocode.SkippedTurnEvent;
import robocode.WinEvent;
import wiki.mc2k7.RaikoGun;

/**
 * BeepBoop - a robot by Kevin Clark (Kev)
 * See http://robowiki.net/wiki/BeepBoop for details.
 */
public class BeepBoop extends AdvancedRobot {
  public static boolean TC = false;
  public static boolean MC = false;
  public static boolean MC2k7 = false;
  public static boolean paint = false;
  public static boolean writeData = false;
  public static boolean verbose = false;

  private static GameState gs;
  private static WaveManager waveManager;
  private static Strategy strategy;
  private static Radar radar;
  private static Shielder shielder;
  private static Gun gun;
  private static Movement movement;
  private static RaikoGun raikoGun;
  private static DatasetWriter datasetWriter;

  private static int[] finishes;
  private static int skippedTurns = 0;

  private ScannedRobotEvent lastScanEvent;
  private Graphics2D lastGraphics;
  private long lastPaintTime;
  private boolean hasWon;

  @Override
  public void run() {
    System.out.println("Beep Boop!");
    setColors(new Color(20, 25, 35), new Color(20, 25, 35), new Color(0, 255, 128));
    setAdjustRadarForGunTurn(true);
    setAdjustGunForRobotTurn(true);

    if(getRoundNum() == 0) {
      if (MC2k7) {
        TC = false;
        MC = true;
        raikoGun = new RaikoGun(this);
      }
      gs = new GameState(this);
      waveManager = new WaveManager();
      strategy = new Strategy(this, waveManager);
      shielder = new Shielder(this, waveManager.enemyWaveManager);
      datasetWriter = new DatasetWriter();
      radar = new Radar(this);
      gun = new Gun(this, strategy, waveManager.myWaveManager, gs.myHitRateTracker, datasetWriter);
      movement = new Movement(
          this, strategy, waveManager.enemyWaveManager, gs.enemyHitRateTracker, datasetWriter);
      finishes = new int[getOthers() + 1];
    }

    BattleField.INSTANCE.init(this);
    gs.init();
    waveManager.init();
    radar.init();
    gun.init();
    movement.init();

    do {
      if ((getOthers() == 0 || radar.getLastScannedTime() < getTime()) && lastScanEvent != null) {
        gs.updateNoScan();
        strategy.update(gs);
        if (!TC && !strategy.shield) {
          Stopwatch.INSTANCE.start("Move");
          movement.move(gs);
          Stopwatch.INSTANCE.stop();
        }
      }
      radar.sweep(gs);
      paint = lastPaintTime == getTime();
      if (paint && lastGraphics != null) {
        Painter.THIS_TICK.paint(lastGraphics);
        Painter.PERMANENT.paint(lastGraphics);
      }
      Painter.THIS_TICK.clear();
      execute();
    } while(true);
  }

  @Override
  public void onScannedRobot(ScannedRobotEvent e) {
    if (writeData) {
      datasetWriter.init(this, e.getName());
    }
    lastScanEvent = e;
    gs.update(e);
    strategy.update(gs);
    radar.scan(gs);
    if (strategy.shield) {
      shielder.shield(gs);
    } else {
      if (!TC) {
        Stopwatch.INSTANCE.start("Move");
        movement.move(gs);
        Stopwatch.INSTANCE.stop();
      }
      if (!MC) {
        Stopwatch.INSTANCE.start("Gun");
        gun.aim(gs);
        Stopwatch.INSTANCE.stop();
      }
      if (MC2k7) {
        raikoGun.onScannedRobot(e);
      }
    }
  }

  @Override
  public void onBulletHit(BulletHitEvent e) {
    if (!hasWon) {
      if (!MC && !strategy.shield) {
        gun.onBulletHit(e, gs);
      }
      gs.onBulletHit(e);
      if (gs.getEnemyState().energy <= 0) {
        gs.myHitRateTracker.logShotPassed(e.getBullet().getPower());
      }
    }
  }

  @Override
  public void onHitByBullet(HitByBulletEvent e) {
    if (!TC && !strategy.shield) {
      movement.onHitByBullet(e, gs);
    }
    gs.onHitByBullet(e);
    if (gs.getMyState().energy <= 0) {
      gs.enemyHitRateTracker.logShotPassed(e.getPower());
    }
  }

  @Override
  public void onBulletHitBullet(BulletHitBulletEvent e) {
    if (strategy.shield) {
      shielder.onBulletHitBullet(e);
    } else {
      if (!TC) {
        movement.onBulletHitBullet(e, gs);
      }
      if (!MC) {
        gun.onBulletHitBullet(e);
      }
      waveManager.onBulletHitBullet();
    }
  }

  @Override
  public void onHitRobot(HitRobotEvent e) {
    gs.onHitRobot(e);
  }

  @Override
  public void onSkippedTurn(SkippedTurnEvent e) {
    skippedTurns++;
    if (gs.gameTime != getTime()) {
      gs.update(lastScanEvent);
    }
    Stopwatch.INSTANCE.stopAll();
  }

  @Override
  public void onWin(WinEvent e) {
    for (Event otherEvent: getAllEvents()) {
      if(otherEvent instanceof BulletHitEvent) {
        onBulletHit((BulletHitEvent)otherEvent);
      }
    }
    hasWon = true;
    onRoundEnd();
  }

  @Override
  public void onDeath(DeathEvent e) {
    if (!hasWon) {
      onRoundEnd();
      System.out.println();
    }
  }

  public void onRoundEnd() {
    finishes[getOthers()]++;
    if (writeData) {
      datasetWriter.onRoundEnd();
    }
    if (!MC) {
      gun.printStats();
    }
    if (!TC) {
      movement.printStats();
    }
    if (verbose) {
      Stopwatch.INSTANCE.print();
    }
    System.out.println("Skipped Turns: " + skippedTurns);
    System.out.print("Finishes: ");
    for (int i = 0; i < finishes.length; i++) {
      System.out.print((i == getOthers() ? "*" : "") + finishes[i] + " ");
    }
    System.out.println();
  }

  @Override
  public void onPaint(Graphics2D g) {
    lastPaintTime = getTime();
    this.lastGraphics = g;
  }

  @Override
  public void setFire(double power) {
    Bullet b = super.setFireBullet(power);
    if (b != null && MC2k7) {
      gs.onMyFire(power);
    }
  }
}
