package kc.mega.game;

import kc.mega.utils.Geom;
import kc.mega.utils.Predict;

import java.awt.geom.Point2D;

public class PredictState {
  public final Point2D.Double location;
  public final double heading;
  public final double velocity;
  public final long gameTime;

  public PredictState(Point2D.Double location, double heading,
      double velocity, long gameTime) {
    this.location = location;
    this.heading = heading;
    this.velocity = velocity;
    this.gameTime = gameTime;
  }

  public PredictState getNextState(int direction, double turn) {
    double nextHeading = heading + Predict.getTurnIncrement(turn, velocity);
    double nextVelocity = Predict.getNextVelocity(velocity, direction);
    return new PredictState(Geom.project(location, nextHeading, nextVelocity),
        nextHeading, nextVelocity, gameTime + 1);
  }

  public PredictState getNextState(double turn) {
    double nextHeading = heading + Predict.getTurnIncrement(turn, velocity);
    return new PredictState(Geom.project(location, nextHeading, velocity),
        nextHeading, velocity, gameTime + 1);
  }

  public double distance(PredictState state) {
    return location.distance(state.location);
  }

  public double absoluteBearing(PredictState state) {
    return Geom.absoluteBearing(location, state.location);
  }
}
