package kc.mega.learning;

import kc.mega.other.Stopwatch;
import kc.mega.wave.WaveWithFeatures;

import java.util.Arrays;
import java.util.List;

import ags.utils.KdTree.Entry;

public class KNNHitRangeModel extends KNNModel<double[]> {
  public static class Builder extends KNNModel.Builder<double[]> {
    public Builder(String name) {
      super(name);
    }

    @Override
    public KNNHitRangeModel build() {
      return new KNNHitRangeModel(this);
    }
  }

  private  KNNHitRangeModel(Builder builder) {
    super(builder);
  }

  @Override
  public void train(WaveWithFeatures w) {
    tree.addPoint(embed(w), w.hitGFRange());
  }

  @Override
  public double getAimAngle(WaveWithFeatures w) {
    return getAimAngle(w, getNumNeighbors());
  }

  @Override
  public double getAimAngleFast(WaveWithFeatures w) {
    return getAimAngle(w, Math.min(maxNeighborsFast, getNumNeighbors()));
  }

  private double getAimAngle(WaveWithFeatures w, int numNeighbors) {
    Stopwatch.INSTANCE.start("Get neighbors");
    List<Entry<double[]>> neighbors = tree.nearestNeighbor(embed(w), numNeighbors, false);
    Stopwatch.INSTANCE.stop();

    if (neighbors.size() == 0) {
      return w.absoluteBearing;
    }

    Stopwatch.INSTANCE.start("Get angle");
    double[] weights = getWeights(neighbors);
    HitEdge[] edges = new HitEdge[2 * neighbors.size()];
    for (int i = 0; i < neighbors.size(); i++) {
      double[] gfRange = neighbors.get(i).value;
      edges[2 * i] = new HitEdge(gfRange[0], weights[i]);
      edges[1 + 2 * i] = new HitEdge(gfRange[1], -weights[i]);
    }
    Arrays.sort(edges);
    double height = 0;
    double bestHeight = 0;
    int bestIndex = 0;
    for (int i = 0; i < edges.length; i++) {
      height += edges[i].weight;
      if (height > bestHeight) {
        bestHeight = height;
        bestIndex = i;
      }
    }
    Stopwatch.INSTANCE.stop();

    double GF = (edges[bestIndex].GF + edges[bestIndex + 1].GF) / 2;
    return w.absoluteBearing + w.getOffset(GF);
  }

  public static class HitEdge implements Comparable<HitEdge> {
    public double GF;
    public double weight;

    public HitEdge(double GF, double weight) {
      this.GF = GF;
      this.weight = weight;
    }

    @Override
    public int compareTo(HitEdge gfScore){
      return (int)Math.signum(GF - gfScore.GF);
    }
  }
}
