package kc.mega.learning;

import kc.mega.game.BattleField;
import kc.mega.game.PredictState;
import kc.mega.utils.KUtils;
import kc.mega.wave.WaveWithFeatures;

import jk.math.FastTrig;

public class SimpleModels {
  public static abstract class SingleAngleModel extends Model {
    private final GFBins bins;

    public SingleAngleModel(String name, GFBins bins) {
      super(name);
      this.bins = bins;
    }

    @Override
    public double[] getDangers(WaveWithFeatures w) {
      double[] dangers = new double[bins.nBins];
      dangers[bins.hitBin(w.getGF(getAimAngle(w) - w.absoluteBearing))] = 1;
      return dangers;
    }
  }

  public static class HOTModel extends SingleAngleModel {
    public HOTModel(GFBins bins) {
      super("HOT", bins);
    }

    @Override
    public double getAimAngle(WaveWithFeatures w) {
      return w.prevAbsoluteBearing;
    }
  }

  public static class CurrentGFModel extends Model {
    private final GFBins bins;

    public CurrentGFModel(GFBins bins) {
      super("CurrentGF");
      this.bins = bins;
    }

    @Override
    public double[] getDangers(WaveWithFeatures w) {
      double[] dangers = new double[bins.nBins];
      dangers[bins.hitBin(w.currentGF)] = 1;
      dangers[bins.hitBin(-w.currentGF)] = 0.5;
      return dangers;
    }
  }

  public static class SimpleTargetingModel extends SingleAngleModel {
    private final boolean isCircular;
    private final boolean stopAtWall;
    private final boolean isAveraged;

    public SimpleTargetingModel(String name, GFBins bins, boolean isCircular,
        boolean stopAtWalls, boolean isAveraged) {
      super(name, bins);
      this.isCircular = isCircular;
      this.stopAtWall = stopAtWalls;
      this.isAveraged = isAveraged;
    }

    @Override
    public double getAimAngle(WaveWithFeatures w) {
      double turnRate = isCircular ? FastTrig.normalRelativeAngle(
          w.enemyState.heading - w.prevEnemyState.heading) : 0;
      PredictState predictedState = w.enemyState.asPredictState();
      if (isAveraged) {
        predictedState = new PredictState(predictedState.location, predictedState.heading,
            w.avgVelocity, predictedState.gameTime);
      }
      int ticks = 0;
      while (++ticks * KUtils.bulletSpeed(w.power) < predictedState.distance(w.myState)) {
        predictedState = predictedState.getNextState(turnRate);
        if (stopAtWall) {
          predictedState = BattleField.INSTANCE.wallCollisionCheck(predictedState, false);
        }
      }
      return w.myState.absoluteBearing(predictedState);
    }
  }
}
