package kc.mega.other;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import robocode.AdvancedRobot;
import robocode.RobocodeFileOutputStream;

public class DatasetWriter {
  public static final float END_ROUND_CODE = 123456;

  private AdvancedRobot bot;
  private String enemyName = "";
  private String timestamp = "";
  private Map<String, RobocodeFileOutputStream> outStreams = new HashMap<>();

  public void init(AdvancedRobot bot, String name) {
    if (enemyName.isEmpty()) {
      System.out.println("WRITING DATA");
      this.bot = bot;
      enemyName = name.replace(" ", "-");
      timestamp = new SimpleDateFormat("MM-dd-HH:mm:ss").format(new Date());
    }
  }

  public void write(String filename, double[] data) {
    Stopwatch.INSTANCE.start("Write");
    String fname = enemyName + "_" + timestamp + "_" + filename;
    try {
      if (!outStreams.containsKey(fname)) {
        File f = bot.getDataFile(fname);
        String filepath = f.getAbsolutePath();
        System.out.println("New file: " + filepath);
        outStreams.put(fname, new RobocodeFileOutputStream(f));
      }
      RobocodeFileOutputStream outStream = outStreams.get(fname);
      outStream.write(encodeDoubleArray(data));
    } catch (IOException e) {
      System.out.println("Error writing to file!");
    }
    Stopwatch.INSTANCE.stop();
  }

  private static byte[] encodeDoubleArray(double[] x) {
    byte[] bytes = new byte[4 * x.length];
    java.nio.ByteBuffer buffer = java.nio.ByteBuffer.wrap(bytes);
    for (int i = 0; i < x.length; i++) {
      for (int j = 0; j < 4; j++) {
        buffer.putFloat(4 * i, (float)x[i]);
      }
    }
    return bytes;
  }

  public void onRoundEnd() {
    for (String fname : outStreams.keySet()) {
      write(fname.split("_")[2], new double[] {DatasetWriter.END_ROUND_CODE});
    }
  }
}
