package kc.mega.other;

import kc.mega.utils.Geom;
import kc.mega.utils.KUtils;

public class HitRateTracker {
  public static final int PRIOR_HITS = 1;
  public static final int PRIOR_SHOTS = 10;

  private int hits = 0;
  private int shots = 0;
  private int shotsPassed = 0;
  private double damage = 0;
  private double damageThisRound = 0;
  private double totalMAE = 0;

  public void init() {
    damageThisRound = 0;
  }

  public String getHitRateStr() {
    return String.format("%d/%d = %.2f +- %.2f", hits, shotsPassed,
        100 * getTrueHitRate(), 100 * hitRateCIWidth());
  }

  public double estimateHitRate(double power) {
    double hitRate = ((double)hits + PRIOR_HITS) / (shotsPassed + PRIOR_SHOTS);
    return KUtils.clip(hitRate * (shotsPassed == 0 ? 1 : totalMAE / shotsPassed) /
        Geom.simpleMaxEscapeAngle(KUtils.bulletSpeed(power)), 0.001, 0.999);
  }

  public double getTrueHitRate() {
    return hits / Math.max(1, (double)shotsPassed);
  }

  public boolean hitRateInBounds(double min, double max) {
    return min <= getMaxHitRate() && max >= getMinHitRate();
  }

  public double getMinHitRate() {
    return Math.max(getTrueHitRate() - hitRateCIWidth(), 0);
  }

  public double getMaxHitRate() {
    return Math.min(getTrueHitRate() + hitRateCIWidth(), 1);
  }

  private double hitRateCIWidth() {
    double z;
    if (shotsPassed < 10) {
      z = 1.96;
    } else if (shotsPassed < 20) {
      z = 1.645;
    } else if (shotsPassed < 50) {
      z = 1.282;
    } else {
      z = 0.842;
    }
    double p = getTrueHitRate();
    return z * Math.sqrt(p * (1 - p) / Math.max(1, shotsPassed));
  }

  public double getDamage() {
    return damage;
  }

  public double getDamageThisRound() {
    return damageThisRound;
  }

  public int getShots() {
    return shots;
  }

  public void logShot() {
    shots++;
  }

  public void logShotPassed(double power) {
    shotsPassed++;
    totalMAE += Geom.simpleMaxEscapeAngle(KUtils.bulletSpeed(power));
  }

  public void logHit(double power, double enemyEnergy) {
    hits++;
    damage += KUtils.bulletDamage(power, enemyEnergy);
    damageThisRound += KUtils.bulletDamage(power, enemyEnergy);
  }
}
