package kc.mega.other;

import kc.mega.BeepBoop;
import kc.mega.game.BattleField;
import kc.mega.game.BotState;
import kc.mega.game.GameState;
import kc.mega.learning.GFBins;
import kc.mega.movement.MovementWave;
import kc.mega.utils.Geom;
import kc.mega.wave.WaveManager;

import java.awt.geom.Point2D;

import jk.math.FastTrig;
import robocode.AdvancedRobot;
import robocode.BulletHitBulletEvent;

public class Shielder {
  private static final int N_BINS = 501;
  private final AdvancedRobot bot;
  private final WaveManager.EnemyWaveManager waveManager;
  private final GFBins bins;

  public Shielder(AdvancedRobot bot, WaveManager.EnemyWaveManager waveManager) {
    this.bot = bot;
    this.waveManager = waveManager;
    this.bins = new GFBins(N_BINS);
  }

  public void shield(GameState gs) {
    if (gs.enemyFiredLastTick) {
      waveManager.add((MovementWave)new MovementWave.Builder()
        .isVirtual(false)
        .bins(bins)
        .myHistory(gs.enemyHistory.subList(1, gs.enemyHistory.size()))
        .enemyHistory(gs.myHistory.subList(1, gs.myHistory.size()))
        .power(gs.lastEnemyBulletPower)
        .hasBullet(gs.enemyFiredLastTick)
        .doPaint(BeepBoop.paint)
        .build());
      for (int i = 1; i < 5; i++) {
        System.out.println(gs.myHistory.get(i).velocity);
      }
    }

    if ((3 / bot.getGunCoolingRate()) - gs.gameTime > 6) {
      position(gs.getMyState(), gs.getEnemyState());
      return;
    }

    bot.setMaxVelocity(0);
  }

  private void position(BotState myState, BotState enemyState) {
    double absoluteBearing = myState.absoluteBearing(enemyState);
    Point2D.Double center = BattleField.INSTANCE.getCenter();
    double targetHeading = 0;
    double bestValue = Double.NEGATIVE_INFINITY;
    for(double heading = 0; heading < 2 * Math.PI; heading += (2 * Math.PI) / 180) {
      Point2D.Double point = Geom.project(myState.location, heading, 8);
      double value =
          Math.log(point.distance(enemyState.location)) +
          0.5 * Math.log(center.distance(0, 0) - point.distance(center)) +
          Math.log(BattleField.INSTANCE.wallDistance(point)) +
          0.005 * Math.abs(FastTrig.sin(FastTrig.normalRelativeAngle(absoluteBearing - heading)));
      if (value > bestValue) {
        bestValue = value;
        targetHeading = heading;
      }
    }
    double turn = FastTrig.normalRelativeAngle(targetHeading - myState.heading);
    int direction = 1;
    if (turn < -Math.PI/2) {
      turn += Math.PI;
      direction = -1;
    } else if (turn > Math.PI/2) {
      turn -= Math.PI;
      direction = -1;
    }
    bot.setTurnRightRadians(turn);
    bot.setAhead(direction * 1000);
    bot.setMaxVelocity(BattleField.INSTANCE.getFieldRectangle().contains(
        Geom.project(myState.location, myState.heading, direction * 50))
        && bestValue < 14 ? 8 : 0);
  }

  public void onBulletHitBullet(BulletHitBulletEvent e) {
    MovementWave hitWave = waveManager.getBulletHitBulletWave(
        new Point2D.Double(e.getHitBullet().getX(), e.getHitBullet().getY()),
        e.getHitBullet().getPower(), bot.getTime());
    hitWave.hitOffset = FastTrig.normalRelativeAngle(
        e.getHitBullet().getHeadingRadians() - hitWave.absoluteBearing);
  }
}
