package kc.mega.other;

import java.util.Optional;

import jk.math.FastTrig;
import kc.mega.game.BattleField;
import kc.mega.game.BotState;
import kc.mega.game.GameState;
import kc.mega.game.PredictState;
import kc.mega.gun.GunWave;
import kc.mega.utils.Geom;
import kc.mega.wave.WaveManager;
import robocode.AdvancedRobot;

public class Strategy {
  private final AdvancedRobot bot;
  private final WaveManager waveManager;

  public boolean walkingStickSmooth;
  public boolean ram;
  public boolean antiRam;
  public boolean antiMirror;
  public double mirrorDistance;
  public int mirrorTicks;
  public boolean shield = false;

  public Strategy(AdvancedRobot bot, WaveManager waveManager) {
    this.bot = bot;
    this.waveManager = waveManager;
  }

  public void update(GameState gs) {
    BotState myState = gs.getMyState(), enemyState = gs.getEnemyState();
    walkingStickSmooth = gs.enemyHitRateTracker.hitRateInBounds(0.001, 0.05);
    boolean ramm = enemyState.energy < 0.03 && myState.energy > 3 &&
        waveManager.enemyWaveManager.getSurfableWaves(myState).size() == 0
        && gs.inactivityTicks < 350;
    if (!ram && ramm && bot.getOthers() > 0) {
      System.out.println("BEEEEEEEEEEEEEEP!");
    }
    ram = ramm;
    double advVel = enemyState.velocity * FastTrig.cos(
        enemyState.heading - myState.absoluteBearing(enemyState));
    antiRam = !ram && myState.distance(enemyState) < (Math.max(advVel, 0) + 8) * 20 &&
        gs.enemyHitRateTracker.hitRateInBounds(0.15, 1);
    if ((gs.gameTime > 200 && gs.enemyHitRateTracker.getShots() == 0) ||
        (antiRam && gs.enemyHitRateTracker.getDamage() > 20) ||
        gs.enemyHitRateTracker.getDamage() > 50) {
      shield = false;
    }
    if (gs.gameTime > 20) {
      mirrorDistance = 0.002 * enemyState.location.distance(
          BattleField.INSTANCE.mirrorLocation(myState.location)) + 0.998 * mirrorDistance;
      mirrorTicks++;
      boolean antiMirrorr = mirrorDistance / (1 - Math.pow(0.998, mirrorTicks)) < 100;
      if (antiMirrorr && !antiMirror) {
        System.out.println("Anti-mirror on!");
      }
      antiMirror = antiMirrorr;
      walkingStickSmooth |= antiMirror;
    }
  }

  public double getAntiMirrorDanger(PredictState myState, double bulletDanger) {
    if (antiMirror) {
      PredictState enemyState = new PredictState(
          BattleField.INSTANCE.mirrorLocation(myState.location),
          0, 0, myState.gameTime);
      Optional<GunWave> o = waveManager.myWaveManager.closestWave(enemyState, true);
      if (o.isPresent()) {
       GunWave w = o.get();
       double bulletGF = w.getGF(w.bullet.heading - w.absoluteBearing);
       double enemyGF = w.getGF(Geom.offset(w.source, enemyState.location, w.absoluteBearing));
       return Math.max(-bulletDanger * 0.75, -Math.exp(-20 * Math.abs(bulletGF - enemyGF)) /
           (1 + Math.max(w.getTicksUntilBreak(enemyState), 0)));
      }
    }
    return 0;
  }
}
