package kc.mega.utils;

import kc.mega.game.BattleField;
import kc.mega.game.BotState;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

public class Predict {
  public static double[] preciseMaxEscapeAngles(
      Point2D.Double source, BotState botState, double bulletSpeed, boolean smooth) {
    List<Point2D.Double> testPoints = new ArrayList<>();
    testPoints.add(highestOffsetReachablePoint(source, botState, bulletSpeed, 1, smooth));
    testPoints.add(highestOffsetReachablePoint(source, botState, bulletSpeed, -1, smooth));
    return Geom.getAbsValuedOffsetRange(
        source, botState.location, Geom.getBotCornerPoints(testPoints));
  }

  public static Point2D.Double highestOffsetReachablePoint(
      Point2D.Double source, BotState botState,
      double bulletSpeed, int direction,  boolean smooth) {
    double absoluteBearing = Geom.absoluteBearing(source, botState.location);
    int orbitDirection = KUtils.orbitDirection(absoluteBearing, botState.heading, direction);
    double velocity = botState.velocity;
    double targetHeading = absoluteBearing + orbitDirection * direction * Math.PI / 2;
    double radius = 0;
    double totalDistance = 0;

    Point2D.Double projectedLocation = botState.location;
    Point2D.Double bestPoint = botState.location;
    double largestOffset = 0;

    while (radius < source.distance(projectedLocation) ||
           radius < Geom.getMaxDistanceToBot(source, projectedLocation)) {
      double offset = orbitDirection * Geom.offset(source, projectedLocation, absoluteBearing);
      if (offset > largestOffset) {
        largestOffset = offset;
        bestPoint = projectedLocation;
      }
      radius += bulletSpeed;
      velocity = getNextVelocity(velocity, direction);
      totalDistance += velocity;
      double smoothedHeading = targetHeading;
      if (smooth) {
        smoothedHeading = BattleField.INSTANCE.walkingStickSmooth(
            botState.location, targetHeading, direction, KUtils.orbitDirection(
                absoluteBearing, targetHeading, KUtils.nonzeroSign(totalDistance)),
            Math.abs(totalDistance), 17.999);
      }
      projectedLocation = Geom.project(botState.location, smoothedHeading, totalDistance);
    }
    return bestPoint;
  }

  public static double getNextVelocity(double v, int d) {
    if (d == 0) {
      return v - Math.signum(v) * Math.min(decel(Math.abs(v)), Math.abs(v));
    }
    return KUtils.clip(
        v + d * (Math.signum(v) * d < 0 ? decel(Math.abs(v)) : 1), -8, 8);
  }

  public static double decel(double speed) {
    return KUtils.clip(1 + speed / 2, 1, 2);
  }

  public static double getMaxTurn(double v) {
    return Math.PI/18 - Math.abs(v) * Math.PI/240;
  }

  public static double getTurnIncrement(double t, double v) {
    double max = getMaxTurn(v);
    return KUtils.clip(t, -max, max);
  }
}
